@@ -52,7 +52,7 @@ void copyElement(sdf::ElementPtr _sdf, const tinyxml2::XMLElement *_xml)
 
 /////////////////////////////////////////////////
 GazeboServer::GazeboServer()
-  : launch::Plugin()
+  : ignition::launch::Plugin()
 {
 }
 
@@ -116,7 +116,7 @@ bool GazeboServer::Load(const tinyxml2::XMLElement *_elem)
 
     // Update compressed file path to name of recording directory path
     std::string cmpPath = std::string(recordPathMod);
-    if (!std::string(1, cmpPath.back()).compare(common::separator("")))
+    if (!std::string(1, cmpPath.back()).compare(ignition::common::separator("")))
       // Remove the separator at end of path
       cmpPath = cmpPath.substr(0, cmpPath.length() - 1);
     cmpPath += ".zip";
@@ -132,30 +132,30 @@ bool GazeboServer::Load(const tinyxml2::XMLElement *_elem)
     {
       // Update compressed file path to name of recording directory path
       cmpPath = std::string(recordPathMod);
-      if (!std::string(1, cmpPath.back()).compare(common::separator(
+      if (!std::string(1, cmpPath.back()).compare(ignition::common::separator(
         "")))
         // Remove the separator at end of path
         cmpPath = cmpPath.substr(0, cmpPath.length() - 1);
       cmpPath += ".zip";
 
       // Check if path or compressed file with same prefix exists
-      if (common::exists(recordPathMod) ||
-        common::exists(cmpPath))
+      if (ignition::common::exists(recordPathMod) ||
+        ignition::common::exists(cmpPath))
       {
         // Overwrite if flag specified
         if (overwrite)
         {
           bool recordMsg = false, cmpMsg = false;
           // Remove files before initializing console log files on top of them
-          if (common::exists(recordPathMod))
+          if (ignition::common::exists(recordPathMod))
           {
             recordMsg = true;
-            common::removeAll(recordPathMod);
+            ignition::common::removeAll(recordPathMod);
           }
-          if (common::exists(cmpPath))
+          if (ignition::common::exists(cmpPath))
           {
             cmpMsg = true;
-            common::removeFile(cmpPath);
+            ignition::common::removeFile(cmpPath);
           }
 
           // Create log file before printing any messages so they can be logged
@@ -180,21 +180,21 @@ bool GazeboServer::Load(const tinyxml2::XMLElement *_elem)
         {
           // Remove the separator at end of path
           if (!std::string(1, recordPathMod.back()).compare(
-            common::separator("")))
+            ignition::common::separator("")))
             recordPathMod = recordPathMod.substr(0, recordPathMod.length() - 1);
-          recordPathMod = common::uniqueDirectoryPath(recordPathMod);
+          recordPathMod = ignition::common::uniqueDirectoryPath(recordPathMod);
 
           cmpPath = std::string(recordPathMod);
           // Remove the separator at end of path
           if (!std::string(1, cmpPath.back()).compare(
-            common::separator("")))
+            ignition::common::separator("")))
             cmpPath = cmpPath.substr(0, cmpPath.length() - 1);
           cmpPath += ".zip";
 
           // If compressed file exists, rename again
-          if (common::exists(cmpPath))
+          if (ignition::common::exists(cmpPath))
           {
-            cmpPath = common::uniqueFilePath(recordPathMod, "zip");
+            cmpPath = ignition::common::uniqueFilePath(recordPathMod, "zip");
 
             size_t extIdx = cmpPath.find_last_of(".");
             recordPathMod = cmpPath.substr(0, extIdx);
@@ -211,19 +211,12 @@ bool GazeboServer::Load(const tinyxml2::XMLElement *_elem)
       {
         ignLogInit(recordPathMod, "server_console.log");
       }
-      // TODO(anyone) In Ignition-D, to be moved to outside and after this
-      //   if-else statement, after all ignLogInit() calls have been finalized,
-      //   so that <path> in SDF will always be ignored in favor of logging both
-      //   console logs and LogRecord recordings to common::ignLogDirectory().
-      //   In Blueprint and Citadel, LogRecord will record to <path> if no
-      //   --record-path is specified on command line.
-      serverConfig.SetLogRecordPath(recordPathMod);
-      serverConfig.SetLogIgnoreSdfPath(true);
     }
     else
     {
       ignLogInit(recordPathMod, "server_console.log");
     }
+    serverConfig.SetLogRecordPath(recordPathMod);
 
     bool compress{false};
     const auto *compressElem = elem->FirstChildElement("compress");
@@ -341,6 +334,15 @@ bool GazeboServer::Load(const tinyxml2::XMLElement *_elem)
     serverConfig.AddPlugin({entityName, entityType, file, name, sdf});
   }
 
+  // Set headless rendering
+  elem = _elem->FirstChildElement("headless_rendering");
+  if (elem)
+  {
+    std::string str = elem->GetText();
+    serverConfig.SetHeadlessRendering(str == "1" ||
+        common::lowercase(str) == "true");
+  }
+
   // Create and run the simulation server
   this->server.reset(new gazebo::Server(serverConfig));
   this->server->Run(false, 0, !run);
