@@ -17,8 +17,10 @@
 
 #include <algorithm>
 #include <ignition/common/Console.hh>
+#include <ignition/common/Image.hh>
 #include <ignition/common/Util.hh>
 #include <ignition/msgs.hh>
+#include <ignition/transport/Publisher.hh>
 
 #include "MessageDefinitions.hh"
 #include "WebsocketServer.hh"
@@ -228,19 +230,6 @@ int rootCallback(struct lws *_wsi,
   // std::lock_guard<std::mutex> mainLock(self->mutex);
   switch (_reason)
   {
-    // Filter network connections.
-    case LWS_CALLBACK_FILTER_NETWORK_CONNECTION:
-      // Prevent too many connections.
-      if (self->maxConnections >= 0 &&
-          self->connections.size()+1 > self->maxConnections)
-      {
-        ignerr << "Skipping new connection, limit of "
-          << self->maxConnections << " has been reached\n";
-        // Return non-zero to close the connection.
-        return -1;
-      }
-      break;
-
     // Open connections.
     case LWS_CALLBACK_ESTABLISHED:
       igndbg << "LWS_CALLBACK_ESTABLISHED\n";
@@ -265,6 +254,7 @@ int rootCallback(struct lws *_wsi,
     case LWS_CALLBACK_SERVER_WRITEABLE:
       {
         std::lock_guard<std::mutex> lock(self->connections[fd]->mutex);
+
         if (!self->connections[fd]->buffer.empty())
         {
           int msgSize = self->connections[fd]->len.front();
@@ -296,9 +286,28 @@ int rootCallback(struct lws *_wsi,
 
     // Handle incoming messages
     case LWS_CALLBACK_RECEIVE:
-      igndbg << "LWS_CALLBACK_RECEIVE\n";
-      self->OnMessage(fd, std::string((const char *)_in));
-      break;
+      {
+        igndbg << "LWS_CALLBACK_RECEIVE\n";
+
+        // Prevent too many connections.
+        if (self->maxConnections >= 0 &&
+            self->connections.size()+1 > self->maxConnections)
+        {
+          ignerr << "Skipping new connection, limit of "
+            << self->maxConnections << " has been reached\n";
+
+          // This will return an error code of 1008 with a reason of
+          // "max_connections".
+          std::string reason = "max_connections";
+          lws_close_reason(_wsi, LWS_CLOSE_STATUS_POLICY_VIOLATION,
+            reinterpret_cast<unsigned char *>(reason.data()), reason.size());
+
+          // Return non-zero to close the connection.
+          return -1;
+        }
+        self->OnMessage(fd, std::string((const char *)_in).substr(0, _len));
+        break;
+      }
 
     default:
       // Do nothing on default.
@@ -358,7 +367,6 @@ bool WebsocketServer::Load(const tinyxml2::XMLElement *_elem)
   }
   this->publishPeriod = std::chrono::duration_cast<std::chrono::nanoseconds>(
       std::chrono::duration<double>(1.0 / hz));
-
   // Get the authorization key, if present.
   elem = _elem->FirstChildElement("authorization_key");
   if (elem)
@@ -404,13 +412,62 @@ bool WebsocketServer::Load(const tinyxml2::XMLElement *_elem)
     }
     catch (...)
     {
-      ignerr << "Failed to convert port[" << elem->GetText() << "] to integer."
-        << std::endl;
+      ignerr << "Failed to convert max_connections[" << elem->GetText()
+        << "] to integer." << std::endl;
     }
     igndbg << "Using maximum connection count of "
       << this->maxConnections << std::endl;
   }
 
+  // Get the msg count per connection.
+  elem = _elem->FirstChildElement("queue_size_per_connection");
+  if (elem)
+  {
+    int size = -1;
+    auto result = elem->QueryIntText(&size);
+    if (result == tinyxml2::XML_SUCCESS && size >= 0)
+    {
+      this->queueSizePerConnection = size;
+    }
+    else
+    {
+      ignerr << "Failed to parse queue_size_per_connection["
+        << elem->GetText() << "]." << std::endl;
+    }
+    igndbg << "Using connection msg queue size of "
+      << this->queueSizePerConnection << std::endl;
+  }
+
+  // Get the msg type subscription limit
+  elem = _elem->FirstChildElement("subscription_limit_per_connection");
+  if (elem)
+  {
+    auto childElem = elem->FirstChildElement("subscription");
+    while (childElem)
+    {
+      auto msgTypeElem = childElem->FirstChildElement("msg_type");
+      auto limitElem = childElem->FirstChildElement("limit");
+      if (msgTypeElem && limitElem)
+      {
+        std::string msgType = msgTypeElem->GetText();
+        int limit = -1;
+        auto result = limitElem->QueryIntText(&limit);
+        if (result == tinyxml2::XML_SUCCESS && limit >= 0)
+        {
+          this->msgTypeSubscriptionLimit[msgType] = limit;
+          igndbg << "Setting msg type subscription limit[" << msgType
+                 << ", " << limit << "]" << std::endl;
+        }
+        else
+        {
+          ignerr << "Failed to parse subscription limit["
+            << msgType << ", " << limitElem->GetText() << "]." << std::endl;
+        }
+      }
+      childElem = childElem->NextSiblingElement("subscription");
+    }
+  }
+
   std::string sslCertFile = "";
   std::string sslPrivateKeyFile = "";
   elem = _elem->FirstChildElement("ssl");
@@ -525,12 +582,23 @@ void WebsocketServer::QueueMessage(Connection *_connection,
     memcpy(buf.get() + LWS_PRE, _data, _size);
 
     std::lock_guard<std::mutex> lock(_connection->mutex);
-    _connection->buffer.push_back(std::move(buf));
-    _connection->len.push_back(_size);
+    if (_connection->buffer.size() < this->queueSizePerConnection)
+    {
+      _connection->buffer.push_back(std::move(buf));
+      _connection->len.push_back(_size);
 
-    std::scoped_lock<std::mutex> runLock(this->runMutex);
-    this->messageCount++;
-    this->runConditionVariable.notify_all();
+      std::scoped_lock<std::mutex> runLock(this->runMutex);
+      this->messageCount++;
+      this->runConditionVariable.notify_all();
+    }
+    else
+    {
+      static bool warned{false};
+      if (!warned)
+      {
+        ignwarn << "Queue size reached for connection" << std::endl;
+      }
+    }
   }
   else
   {
@@ -593,7 +661,7 @@ void WebsocketServer::OnDisconnect(int _socketId)
 }
 
 //////////////////////////////////////////////////
-void WebsocketServer::OnMessage(int _socketId, const std::string &_msg)
+void WebsocketServer::OnMessage(int _socketId, const std::string _msg)
 {
   // Skip invalid sockets
   if (this->connections.find(_socketId) == this->connections.end())
@@ -607,7 +675,7 @@ void WebsocketServer::OnMessage(int _socketId, const std::string &_msg)
       // Count the number of commas to handle a frame like "sub,,,"
       std::count(_msg.begin(), _msg.end(), ',') != 3)
   {
-    ignerr << "Received an invalid frame with " << frameParts.size()
+    ignerr << "Received an invalid frame[" << _msg << "] with " << frameParts.size()
       << "components when 4 is expected.\n";
     return;
   }
@@ -700,6 +768,36 @@ void WebsocketServer::OnMessage(int _socketId, const std::string &_msg)
     this->QueueMessage(this->connections[_socketId].get(),
         data.c_str(), data.length());
   }
+  else if (frameParts[0] == "topics-types")
+  {
+    igndbg << "Topic and message type list request recieved\n";
+    ignition::msgs::Publishers msg;
+
+    std::vector<std::string> topics;
+
+    // Get the list of topics
+    this->node.TopicList(topics);
+
+    // Store the topics in a message and serialize the message.
+    for (const std::string &topic : topics)
+    {
+      std::vector<transport::MessagePublisher> publishers;
+      this->node.TopicInfo(topic, publishers);
+      for (const transport::MessagePublisher &publisher : publishers)
+      {
+        msgs::Publish *pubMsg = msg.add_publisher();
+        pubMsg->set_topic(topic);
+        pubMsg->set_msg_type(publisher.MsgTypeName());
+      }
+    }
+
+    std::string data = BUILD_MSG(this->operations[PUBLISH], frameParts[0],
+        std::string("ignition.msgs.Publishers"), msg.SerializeAsString());
+
+    // Queue the message for delivery.
+    this->QueueMessage(this->connections[_socketId].get(),
+        data.c_str(), data.length());
+  }
   else if (frameParts[0] == "worlds")
   {
     igndbg << "World info request recieved\n";
@@ -748,19 +846,213 @@ void WebsocketServer::OnMessage(int _socketId, const std::string &_msg)
     this->QueueMessage(this->connections[_socketId].get(),
         data.c_str(), data.length());
   }
+  /// \todo(nkoeng) Deprecate this in Ignition Fortress, and instruct users
+  /// to rely on the "scene" message.
+  else if (frameParts[0] == "particle_emitters")
+  {
+    igndbg << "Particle emitter request received for world["
+      << frameParts[1] << "]\n";
+    ignition::msgs::Empty req;
+    req.set_unused(true);
+
+    ignition::msgs::ParticleEmitter_V rep;
+    bool result;
+    unsigned int timeout = 2000;
+
+    std::string serviceName = std::string("/world/") + frameParts[1] +
+      "/particle_emitters";
+
+    bool executed = this->node.Request(serviceName, req, timeout, rep, result);
+    if (!executed || !result)
+    {
+      ignerr << "Failed to get the particle emitter information for "
+        << frameParts[1] << " world.\n";
+    }
+
+    std::string data = BUILD_MSG(this->operations[PUBLISH], frameParts[0],
+        std::string("ignition.msgs.ParticleEmitter_V"),
+        rep.SerializeAsString());
+
+    // Queue the message for delivery.
+    this->QueueMessage(this->connections[_socketId].get(),
+        data.c_str(), data.length());
+  }
   else if (frameParts[0] == "sub")
   {
+    std::string topic = frameParts[1];
+
+    // check and update subscription count
+    if (!this->UpdateMsgTypeSubscriptionCount(topic, _socketId, true))
+      return;
+
     // Store the relation of socketId to subscribed topic.
-    this->topicConnections[frameParts[1]].insert(_socketId);
-    this->topicTimestamps[frameParts[1]] =
+    this->topicConnections[topic].insert(_socketId);
+    this->topicTimestamps[topic] =
       std::chrono::steady_clock::now() - this->publishPeriod;
 
     igndbg << "Subscribe request to topic[" << frameParts[1] << "]\n";
-    this->node.SubscribeRaw(frameParts[1],
+    this->node.SubscribeRaw(topic,
         std::bind(&WebsocketServer::OnWebsocketSubscribedMessage,
           this, std::placeholders::_1,
           std::placeholders::_2, std::placeholders::_3));
   }
+  else if (frameParts[0] == "image")
+  {
+    std::string topic = frameParts[1];
+
+    // check and update subscription count
+    if (!this->UpdateMsgTypeSubscriptionCount(topic, _socketId, true))
+      return;
+
+    // Store the relation of socketId to subscribed topic.
+    this->topicConnections[topic].insert(_socketId);
+    this->topicTimestamps[topic] =
+      std::chrono::steady_clock::now() - this->publishPeriod;
+
+    std::vector<std::string> allTopics;
+    std::set<std::string> imageTopics;
+    this->node.TopicList(allTopics);
+    for (auto queryTopic: allTopics)
+    {
+      std::vector<transport::MessagePublisher> publishers;
+      this->node.TopicInfo(queryTopic, publishers);
+      for (auto pub: publishers)
+      {
+        if (pub.MsgTypeName() == "ignition.msgs.Image")
+        {
+          imageTopics.insert(queryTopic);
+          break;
+        }
+      }
+    }
+
+    if (!imageTopics.count(topic))
+    {
+      igndbg << "Could not find topic: " << topic  << " to stream"
+                << std::endl;
+      return;
+    }
+
+    igndbg << "Subscribe request to image topic[" << frameParts[1] << "]\n";
+    this->node.Subscribe(frameParts[1],
+        &WebsocketServer::OnWebsocketSubscribedImageMessage, this);
+  }
+  else if (frameParts[0] == "unsub")
+  {
+    std::string topic = frameParts[1];
+
+    igndbg << "Unsubscribe request for topic[" << topic << "]\n";
+    std::map<std::string, std::set<int>>::iterator topicConnectionIter =
+      this->topicConnections.find(topic);
+
+    if (topicConnectionIter != this->topicConnections.end())
+    {
+      // Remove from the topic connections map
+      topicConnectionIter->second.erase(_socketId);
+
+      // remove from the connection's topic throttling maps
+      auto &con = this->connections[_socketId];
+      con->topicPublishPeriods.erase(topic);
+      con->topicTimestamps.erase(topic);
+
+      // check and update subscription count
+      this->UpdateMsgTypeSubscriptionCount(topic, _socketId, false);
+
+      // Only unsubscribe from the Ignition Transport topic if there are no
+      // more websocket connections.
+      if (topicConnectionIter->second.empty())
+      {
+        igndbg << "Unsubscribing from Ignition Transport Topic["
+          << frameParts[1] << "]\n";
+        this->node.Unsubscribe(frameParts[1]);
+      }
+    }
+    else
+    {
+      ignwarn << "The websocket server is not subscribed to topic["
+        << topic << "]. Unable to unsubscribe from the topic\n";
+    }
+  }
+  else if (frameParts[0] == "throttle")
+  {
+    std::string topic = frameParts[1];
+    igndbg << "Throttle request for topic[" << topic << "]\n";
+    if (!topic.empty())
+    {
+      try
+      {
+        int rate = std::stoi(frameParts[3]);
+        double period = 1.0 / static_cast<double>(rate);
+        this->connections[_socketId]->topicPublishPeriods[topic] =
+            std::chrono::duration_cast<std::chrono::nanoseconds>(
+            std::chrono::duration<double>(1.0 / rate));
+      }
+      catch (...)
+      {
+        ignwarn << "Unable to set topic rate for topic[" << topic
+                << "]" << std::endl;
+      }
+    }
+  }
+  else if (frameParts[0] == "asset")
+  {
+    this->OnAsset(_socketId, frameParts);
+  }
+}
+
+//////////////////////////////////////////////////
+void WebsocketServer::OnAsset(int _socketId,
+    const std::vector<std::string> &_frameParts)
+{
+  if (_frameParts.size() <= 1)
+  {
+    ignerr << "Asset requested, but asset URI is missing\n";
+    return;
+  }
+
+  std::string assetUri = _frameParts[1];
+
+  std::string resolvedPath;
+
+  // Short circuit the case where the assetURI is already a valid path.
+  if (common::exists(assetUri))
+  {
+    resolvedPath = assetUri;
+  }
+  else
+  {
+    ignition::msgs::StringMsg req, rep;
+    req.set_data(assetUri);
+    bool result;
+    unsigned int timeout = 2000;
+
+    // Request the file path from Gazebo
+    bool executed = this->node.Request("/gazebo/resource_paths/resolve",
+        req, timeout, rep, result);
+    if (executed && result)
+      resolvedPath = rep.data();
+  }
+
+  if (!resolvedPath.empty())
+  {
+    // Read the file
+    std::ifstream infile(resolvedPath, std::ios_base::binary);
+    std::string fileBuffer = std::string(
+        std::istreambuf_iterator<char>(infile),
+        std::istreambuf_iterator<char>());
+
+    // Store the file in a protobuf message
+    ignition::msgs::Bytes bytes;
+    bytes.set_data(fileBuffer);
+
+    // Construct the response message
+    std::string data = BUILD_MSG(this->operations[ASSET], assetUri,
+        std::string("ignition.msgs.Bytes"), bytes.SerializeAsString());
+
+    // Queue the message for delivery.
+    this->QueueMessage(this->connections[_socketId].get(),
+        data.c_str(), data.length());
+  }
 }
 
 //////////////////////////////////////////////////
@@ -797,6 +1089,92 @@ void WebsocketServer::OnWebsocketSubscribedMessage(
       // Construct the final message.
       std::string msg = header->second + std::string(_data, _size);
 
+      // Send the message
+      for (const int &socketId : iter->second)
+      {
+        auto conIt = this->connections.find(socketId);
+        if (conIt != this->connections.end())
+        {
+          // do additional throttling based on client connection setting
+          auto lastPubTimeCon = conIt->second->topicTimestamps[_info.Topic()];
+          std::chrono::nanoseconds timeDeltaCon = systemTime - lastPubTimeCon;
+          if (timeDeltaCon >= conIt->second->topicPublishPeriods[_info.Topic()])
+          {
+            conIt->second->topicTimestamps[_info.Topic()] = systemTime;
+            this->QueueMessage(conIt->second.get(),
+                msg.c_str(), msg.length());
+          }
+        }
+      }
+    }
+  }
+}
+
+//////////////////////////////////////////////////
+void WebsocketServer::OnWebsocketSubscribedImageMessage(
+    const ignition::msgs::Image &_msg,
+    const ignition::transport::MessageInfo &_info)
+{
+  std::map<std::string, std::set<int>>::const_iterator iter =
+    this->topicConnections.find(_info.Topic());
+
+  if (iter != this->topicConnections.end())
+  {
+    std::lock_guard<std::mutex> mainLock(this->subscriptionMutex);
+    std::chrono::time_point<std::chrono::steady_clock> systemTime =
+      std::chrono::steady_clock::now();
+
+    std::chrono::nanoseconds timeDelta =
+      systemTime - this->topicTimestamps[_info.Topic()];
+
+    if (timeDelta > this->publishPeriod)
+    {
+      // Get the header, or build a new header if it doesn't exist.
+      auto header = this->publishHeaders.find(_info.Topic());
+      if (header == this->publishHeaders.end())
+      {
+        this->publishHeaders[_info.Topic()] = BUILD_HEADER(
+          this->operations[PUBLISH], _info.Topic(), _info.Type());
+        header = this->publishHeaders.find(_info.Topic());
+      }
+
+      // Store the last time this topic was published.
+      this->topicTimestamps[_info.Topic()] = systemTime;
+
+      // convert to RGB image if needed
+      common::Image image;
+      switch (_msg.pixel_format_type())
+      {
+        case msgs::PixelFormatType::RGB_INT8:
+          image.SetFromData(reinterpret_cast<const unsigned char *>(
+              _msg.data().c_str()),
+              _msg.width(), _msg.height(), common::Image::RGB_INT8);
+          break;
+        case msgs::PixelFormatType::R_FLOAT32:
+          common::Image::ConvertToRGBImage<float>(_msg.data().c_str(),
+              _msg.width(), _msg.height(), image,
+              0, std::numeric_limits<float>::lowest(), true);
+          break;
+        case msgs::PixelFormatType::L_INT16:
+          common::Image::ConvertToRGBImage<uint16_t>(_msg.data().c_str(),
+              _msg.width(), _msg.height(), image);
+          break;
+        case msgs::PixelFormatType::L_INT8:
+          common::Image::ConvertToRGBImage<uint8_t>(_msg.data().c_str(),
+              _msg.width(), _msg.height(), image);
+          break;
+        default:
+          return;
+      }
+
+      // alway publish rgb_int8 format
+      std::vector<unsigned char> buffer;
+      image.SavePNGToBuffer(buffer);
+      std::string img(reinterpret_cast<char *>(buffer.data()), buffer.size());
+
+      // Construct the final message.
+      std::string msg = header->second + img;
+
       // Send the message
       for (const int &socketId : iter->second)
       {
@@ -809,3 +1187,77 @@ void WebsocketServer::OnWebsocketSubscribedMessage(
     }
   }
 }
+
+//////////////////////////////////////////////////
+bool WebsocketServer::UpdateMsgTypeSubscriptionCount(const std::string &_topic,
+    int _socketId, bool _subscribe)
+{
+  // check if limit reached for the subscribed msg type
+  // if not, update subscription count
+  std::vector<transport::MessagePublisher> publishers;
+  this->node.TopicInfo(_topic, publishers);
+  if (!publishers.empty())
+  {
+    std::string msgType = publishers.begin()->MsgTypeName();
+    auto limitIt = this->msgTypeSubscriptionLimit.find(msgType);
+    if (limitIt != this->msgTypeSubscriptionLimit.end())
+    {
+      bool limitReached = false;
+      auto conIt = this->connections.find(_socketId);
+      if (conIt != this->connections.end())
+      {
+        auto &con = conIt->second;
+        auto &subCount = con->msgTypeSubscriptionCount;
+        auto countIt = subCount.find(msgType);
+
+        // if there is already a subscription on the topic for this connection
+        if (countIt != subCount.end())
+        {
+          // subscribe: increment count and check if reached limit
+          // unsubscribe: decrement count and make sure count is >= 0
+          if (_subscribe)
+          {
+            if (countIt->second + 1 <= limitIt->second)
+            {
+              countIt->second++;
+            }
+            else
+            {
+              limitReached = true;
+            }
+          }
+          else
+          {
+            countIt->second = std::max(0, countIt->second - 1);
+          }
+        }
+        // if topic not yet subscribed, set count to 1 on subscription
+        // ignore for unsubscribe option
+        else if (limitIt->second > 0)
+        {
+          if (_subscribe)
+            subCount[msgType] = 1;
+        }
+        // corner case when msg type subscription limit is set to 0
+        else if (_subscribe)
+        {
+          limitReached = true;
+        }
+        if (limitReached)
+        {
+          ignwarn << "Msg type subscription limit reached[" << msgType
+              << ", " << limitIt->second << "] for connection[" << _socketId
+              << "]" << std::endl;
+          return false;
+        }
+      }
+      else
+      {
+        ignwarn << "Unable to find connection[" << _socketId << "]"
+            << " when setting subscription limit." << std::endl;
+        return false;
+      }
+    }
+  }
+  return true;
+}
